#include <windows.h> 
#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h> 

#define FUNC __declspec(dllimport) __stdcall
extern "C" HANDLE FUNC OpenHidHandle(unsigned short vendor_id, unsigned short product_id);
extern "C" void FUNC ReadReport(HANDLE handle,unsigned char *InputReport,int *len);
extern "C" void FUNC WriteReport(HANDLE handle,unsigned char *OutputReport, int *len);
extern "C" void FUNC CloseHidHandle(HANDLE handle);

HANDLE hRsDevHandle;

const USHORT VID = 0x057e; // Nintendo (?)
const USHORT PID = 0x0306; // Wii-Remote (?)

#define MAXREPORTSIZE 256
int		InputLength,OutputLength;
unsigned char	InputReport[MAXREPORTSIZE];
unsigned char	OutputReport[MAXREPORTSIZE];

unsigned char	button1,button2;
unsigned char	Ax,Ay,Az;
unsigned char	wii_remote_loop_flag;
unsigned char	wii_remote_in_use;

int				Ix1,Iy1,Ix2,Iy2;

void Wii_Remote_Input(void);
void Wii_Remote_mode_0x31(void);
int Wii_Remote_check(void);

void Sleep2(int msec)
{
	LARGE_INTEGER f;
	LARGE_INTEGER t1,t2;
	double t;
	QueryPerformanceFrequency(&f);
	QueryPerformanceCounter(&t1);
	while (1) {
		QueryPerformanceCounter(&t2);
		t = (double)t2.QuadPart - (double)t1.QuadPart;
		t /= (double)f.QuadPart;
		t *= 1000;
		if (t>(double)msec) { return; }
	}
}

void Report_0x11(unsigned char data)
{
	OutputReport[0]=0x11; // Report ID
	OutputReport[1]=data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x12(unsigned char data1,unsigned char data2)
{
	OutputReport[0]=0x12; // Report ID
	OutputReport[1]=data1;
	OutputReport[2]=data2;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x13(unsigned char data)
{
	OutputReport[0]=0x13; // Report ID
	OutputReport[1]=data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x16_type1(unsigned char adr,unsigned char len,unsigned char *buf)
{

	OutputReport[0]=0x16; // Report ID
	OutputReport[1]=0x04; // 00(EEPROM) or 0x04(other I2C 1byte type) or 0x08(I2C 2byte type)
	OutputReport[2]=adr; // I2C ADRS
	OutputReport[3]=0xFF; // ignore
	OutputReport[4]=buf[0];  // Mem_Adr or Cmd
	OutputReport[5]=len-1;  // len >= 2
	OutputReport[6]=buf[1]; OutputReport[7]=buf[2]; OutputReport[8]=buf[3]; OutputReport[9]=buf[4];
	OutputReport[10]=buf[5];OutputReport[11]=buf[6];OutputReport[12]=buf[7];OutputReport[13]=buf[8];
	OutputReport[14]=buf[9];OutputReport[15]=buf[10];OutputReport[16]=buf[11];OutputReport[17]=buf[12];
	OutputReport[18]=buf[13];OutputReport[19]=buf[14];OutputReport[20]=buf[15];OutputReport[21]=buf[16];
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x1a(unsigned char data)
{
	OutputReport[0]=0x1a; // Report ID
	OutputReport[1]=data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Wii_Remote_mode_0x30(void)
{
	// Set Report Type
	Report_0x12(0x00,0x30);  // button ONLY Report Mode
}

void Wii_Remote_mode_0x31(void)
{
	// Set Report Type
	Report_0x12(0x00,0x31);  // Acc Sensor + button  Report Mode
}

void Wii_IR_init(void)
{
	unsigned char buf[16];
	unsigned char p0,p1,p2,p3;

	Report_0x13(0x04); // IR on 1  ... OSC 25MHz --> start
	Sleep2(10);
	Report_0x1a(0x04); // IR on 2  ... Chip enable #CE --> Low (Enable)
	
	// Setup IR Sensivity
	// p0=0x72; p1=0x20; p2=0x1F; p3=0x03; // Sensitivity 1
	   p0=0xC8; p1=0x36; p2=0x35; p3=0x03; // 2
	// p0=0xAA; p1=0x64; p2=0x63; p3=0x03; // 3
	// p0=0x96; p1=0xB4; p2=0xB3; p3=0x04; // 4
	// p0=0x96; p1=0xFE; p3=0xFE; p3=0x05; // 5
		
	buf[0]=0x30; buf[1]=0x01; // ??
	Report_0x16_type1(0xB0, 2 , buf); Sleep2(10);
	buf[0]=0x00; buf[1]=0x02; buf[2]=0x00; buf[3]=0x00; buf[4]=0x71; buf[5]=0x01; buf[6]=0x00; buf[7]=p0;
	Report_0x16_type1(0xB0, 8 , buf); Sleep2(10);
	buf[0]=0x07; buf[1]=0x00; buf[2]=p1;
	Report_0x16_type1(0xB0, 3 , buf); Sleep2(10);
	buf[0]=0x1A; buf[1]=p2; buf[2]=p3;
	Report_0x16_type1(0xB0, 3 , buf); Sleep2(10);
	buf[0]=0x33; buf[1]=0x03;
	Report_0x16_type1(0xB0, 2 , buf); Sleep2(10);
	buf[0]=0x30; buf[1]=0x08; // ??
	Report_0x16_type1(0xB0, 2 , buf); Sleep2(10);
	
	Report_0x12(0x00,0x33); // mode 0x33 ... IR Interrupt start
}

void Wii_Remote_Input(void)
{
	unsigned char r;
	int s;
	ReadReport( hRsDevHandle , InputReport, &InputLength);
	
	r = InputReport[0];
	if ( (r & 0xF0) == 0x30 )  {
		button1=InputReport[1]; button2=InputReport[2];
		
		if ( r & 0x01 ) {
			Ax = InputReport[3];
			Ay = InputReport[4];
			Az = InputReport[5];
		}
		if ( r & 0x02 ) {
			// ir
			Ix1 = InputReport[6];
			Iy1 = InputReport[7];
			s   = InputReport[8];
			Ix1 += (s & 0x30) <<4;
			Iy1 += (s & 0xC0) <<2;
			
			Ix2 = InputReport[9];
			Iy2 = InputReport[10];
			s   = InputReport[11];
			Ix2 += (s & 0x30) <<4;
			Iy2 += (s & 0xC0) <<2;
		}
	}
}

void Wii_Remote_Input_Main(void *)
{
	// wii remote input thread
	while (wii_remote_loop_flag) {
		wii_remote_in_use=1;
		Wii_Remote_Input();
		Sleep(100);
	}
	wii_remote_in_use=0;
	return;
}

uintptr_t wii_th;

int Wii_Remote_check(void)
{
	Ax=0;
	wii_remote_loop_flag=1;
	wii_th = _beginthread( Wii_Remote_Input_Main , 0 , NULL );
	Sleep2(100);
	wii_remote_loop_flag=0;
	return (Ax!=0);
}

void Stop_Wii_remote_th(void)
{
	if (wii_remote_in_use==1) {
		TerminateThread( (HANDLE)wii_th , -1); // kill thread  if it stopped
	}
}

void MainLoop(void)
{
	POINT pt;
	int ax,ay;
	int vx,vy;
	int lx,ly;
	int bt1,bt2,bt3,bt4,bt5,bt6;
	int px,py, px_,py_;
	int ir_track;
	
	vx=0;vy=0;
	bt1=0;bt2=0;bt3=0;bt4=0;bt5=0;bt6=0;
	px_=0;py_=0; ir_track= -1;
	
	while (1) {
		if (kbhit()) { if (getch()==0x1B) { break; } } // ESC key  to Quit
		
		Wii_Remote_Input(); // Wii Remote
		if (button2==0x80) { break; }  // Home button  to Quit
		
		if (button2&0x08) { // A button
			if (bt1==0) { mouse_event(MOUSEEVENTF_RIGHTDOWN,0, 0, 0, NULL); bt1=1; }
		} else {
			if (bt1==1) { mouse_event(MOUSEEVENTF_RIGHTUP,0, 0, 0, NULL); bt1=0; }
		}
		if (button2&0x04) { // B button
			if (bt2==0) { mouse_event(MOUSEEVENTF_LEFTDOWN,0, 0, 0, NULL); bt2=1; }
		} else {
			if (bt2==1) { mouse_event(MOUSEEVENTF_LEFTUP,0, 0, 0, NULL); bt2=0; }
		}
		if (button2&0x02) { // (1) button
			if (bt3==0) { bt3=1; }
			mouse_event(MOUSEEVENTF_WHEEL,0, 0, 64, 0);
		} else {
			bt3=0;
		}
		if (button2&0x01) { // (2) button
			if (bt4==0) { bt4=1; }
			mouse_event(MOUSEEVENTF_WHEEL,0, 0, -64, 0);
		} else {
			bt4=0;
		}
		
		if (button1&0x10) {  // (+) button
			if (bt5==0) {
				Report_0x13(0x05); Sleep2(80); Report_0x13(0x04);  //Rumble
				mouse_event(MOUSEEVENTF_WHEEL,0, 0, -64, 0);
				bt5=1;
			}
		} else {
			bt5=0;
		}
		if (button2&0x10) {  // (-) button
			if (bt6==0) {
				Report_0x13(0x05); Sleep2(80); Report_0x13(0x04);  //Rumble
				mouse_event(MOUSEEVENTF_WHEEL,0, 0, 64, 0);
				bt6=1;
			}
		} else {
			bt6=0;
		}
		
		ax=0; ay=0;
		if (button1&0x08) { ay = 1;  } // up
		if (button1&0x04) { ay = -1; } // down
		if (button1&0x02) { ax = 1;  } // right
		if (button1&0x01) { ax = -1; } // left
		
		
		if ((Ix1!=1023)&&(Ix2!=1023)) { // 2-IR-data OK
			if (ir_track==0) {
				ir_track=1; printf("IR-Track Ok\n");
			}
			px= (Ix1 + Ix2) /2;  px = -(px - 512) * 3 + 512;
			py= (Iy1 + Iy2) /2;  py =  (py - 512) * 3 + 512;
			if ((px_==0)&&(py_==0)) { px_=px; py_=py; }
			px = (px + px_*5)/6;  py = (py + py_*5)/6; // filter
			SetCursorPos(px,py);
			px_ = px; py_= py;
		} else {
			if (ir_track!=0) {
				ir_track=0; printf("IR-Track NG\n");
			}
		}
		
		
		// move Mouse Cursor 
		GetCursorPos(&pt);
		lx = pt.x; ly = pt.y;
		if ( (ax>0)&&(vx>=0) ) { vx++; if (vx>  16) { vx=  16;}  }
		if ( (ax<0)&&(vx<=0) ) { vx--; if (vx< -16) { vx= -16;}  }
		if ( (ax<0)&&(vx>=0) ) { vx/=4; }
		if ( (ax>0)&&(vx<=0) ) { vx/=4; }
		if ( (ay>0)&&(vy>=0) ) { vy++; if (vx>  16) { vx=  16;}  }
		if ( (ay<0)&&(vy<=0) ) { vy--; if (vx< -16) { vx= -16;}  }
		if ( (ay<0)&&(vy>=0) ) { vy/=4; }
		if ( (ay>0)&&(vy<=0) ) { vy/=4; }
		if ( ax == 0 ) { vx /=2; }
		if ( ay == 0 ) { vy /=2; }
		if (vx >  2) { lx += (vx-2); }
		if (vx < -2) { lx += (vx+2); }
		if (vy >  2) { ly -= (vy-2); }
		if (vy < -2) { ly -= (vy+2); }
		SetCursorPos(lx,ly);
	}
}

int main( int argc, char *argv[])
{ 
	int i;
	unsigned char buf[16];
	unsigned char p0,p1,p2,p3;
	
	hRsDevHandle = OpenHidHandle(VID,PID);
	if ( hRsDevHandle == INVALID_HANDLE_VALUE) {
		MessageBox(0,"Wii-Remote not found.",0,0);
		return 1;
	}
	
	Wii_Remote_mode_0x31();
	if ( Wii_Remote_check() ) {
		printf("###########################################\n");
		printf("#                                         #\n");
		printf("#   Wii-Pointer Test version 0.2          #\n");
		printf("#                  build 2007.2.17        #\n");
		printf("#                                         #\n");
		printf("###########################################\n");
		printf("\n");
		printf("Usage:\n");
		printf("   up/down/left/right button  ... Move cursor\n");
		printf("   B button  ... Left click\n");
		printf("   A button  ... Right click\n");
		printf("   (1) button  ... Wheel up\n");
		printf("   (2) button  ... Wheel down\n");
		printf("   (+) button  ... Wheel down\n");
		printf("   (-) button  ... Wheel up\n");
		printf("   Home button  ... Quit\n");
		printf("\n");

		Report_0x11(0x10); // LED1 on
		Wii_IR_init();

		MainLoop();
		
		Report_0x13(0x00); // IR off 1
		Report_0x1a(0x00); // IR off 2
		Wii_Remote_mode_0x30();
				
	} else {
		
		MessageBox(0,"Wii-Remote cannot open.",0,0);
		Stop_Wii_remote_th();
	}
	CloseHidHandle( hRsDevHandle );
	return 0;
} 
